@@ -1,63 +1,49 @@
 from __future__ import annotations
 
 import re
-from typing import (
-    TYPE_CHECKING,
-    Any,
-    Callable,
-    ClassVar,
-    Iterable,
-    Iterator,
-    Literal,
-    Mapping,
-    Sequence,
-    TypeVar,
-)
+from itertools import chain
+from typing import TYPE_CHECKING, Any, Callable, ClassVar, Protocol, TypeVar
 
 from narwhals._compliant.typing import (
     CompliantDataFrameAny,
-    CompliantDataFrameT,
     CompliantDataFrameT_co,
     CompliantExprT_contra,
     CompliantFrameT,
     CompliantFrameT_co,
     CompliantLazyFrameAny,
-    CompliantLazyFrameT,
     DepthTrackingExprAny,
     DepthTrackingExprT_contra,
     EagerExprT_contra,
-    LazyExprT_contra,
-    NativeExprT_co,
+    NarwhalsAggregation,
 )
-from narwhals._typing_compat import Protocol38
-from narwhals.utils import is_sequence_of
+from narwhals._utils import is_sequence_of
 
 if TYPE_CHECKING:
-    from typing_extensions import TypeAlias
+    from collections.abc import Iterable, Iterator, Mapping, Sequence
+
+    from narwhals._compliant.expr import CompliantExpr
 
     _SameFrameT = TypeVar("_SameFrameT", CompliantDataFrameAny, CompliantLazyFrameAny)
 
 
-__all__ = [
-    "CompliantGroupBy",
-    "DepthTrackingGroupBy",
-    "EagerGroupBy",
-    "LazyGroupBy",
-    "NarwhalsAggregation",
-]
+__all__ = ["CompliantGroupBy", "DepthTrackingGroupBy", "EagerGroupBy"]
 
 NativeAggregationT_co = TypeVar(
     "NativeAggregationT_co", bound="str | Callable[..., Any]", covariant=True
 )
-NarwhalsAggregation: TypeAlias = Literal[
-    "sum", "mean", "median", "max", "min", "std", "var", "len", "n_unique", "count"
-]
 
 
 _RE_LEAF_NAME: re.Pattern[str] = re.compile(r"(\w+->)")
 
 
-class CompliantGroupBy(Protocol38[CompliantFrameT_co, CompliantExprT_contra]):
+def _evaluate_aliases(
+    frame: CompliantFrameT, exprs: Iterable[CompliantExpr[CompliantFrameT, Any]], /
+) -> list[str]:
+    it = (expr._evaluate_aliases(frame) for expr in exprs)
+    return list(chain.from_iterable(it))
+
+
+class CompliantGroupBy(Protocol[CompliantFrameT_co, CompliantExprT_contra]):
     _compliant_frame: Any
 
     @property
@@ -78,25 +64,24 @@ class CompliantGroupBy(Protocol38[CompliantFrameT_co, CompliantExprT_contra]):
 
 class DataFrameGroupBy(
     CompliantGroupBy[CompliantDataFrameT_co, CompliantExprT_contra],
-    Protocol38[CompliantDataFrameT_co, CompliantExprT_contra],
+    Protocol[CompliantDataFrameT_co, CompliantExprT_contra],
 ):
     def __iter__(self) -> Iterator[tuple[Any, CompliantDataFrameT_co]]: ...
 
 
 class ParseKeysGroupBy(
-    CompliantGroupBy[CompliantFrameT, CompliantExprT_contra],
-    Protocol38[CompliantFrameT, CompliantExprT_contra],
+    CompliantGroupBy[CompliantFrameT_co, CompliantExprT_contra],
+    Protocol[CompliantFrameT_co, CompliantExprT_contra],
 ):
     def _parse_keys(
         self,
-        compliant_frame: CompliantFrameT,
+        compliant_frame: _SameFrameT,
         keys: Sequence[CompliantExprT_contra] | Sequence[str],
-    ) -> tuple[CompliantFrameT, list[str], list[str]]:
+    ) -> tuple[_SameFrameT, list[str], list[str]]:
         if is_sequence_of(keys, str):
             keys_str = list(keys)
             return compliant_frame, keys_str, keys_str.copy()
-        else:
-            return self._parse_expr_keys(compliant_frame, keys=keys)
+        return self._parse_expr_keys(compliant_frame, keys=keys)
 
     @staticmethod
     def _parse_expr_keys(
@@ -123,7 +108,7 @@ class ParseKeysGroupBy(
             key_str = str(key)  # pandas allows non-string column names :sob:
             return f"_{key_str}_tmp{'_' * (tmp_name_length - len(key_str) - 5)}"
 
-        output_names = compliant_frame._evaluate_aliases(*keys)
+        output_names = _evaluate_aliases(compliant_frame, keys)
 
         safe_keys = [
             # multi-output expression cannot have duplicate names, hence it's safe to suffix
@@ -135,14 +120,14 @@ class ParseKeysGroupBy(
         ]
         return (
             compliant_frame.with_columns(*safe_keys),
-            compliant_frame._evaluate_aliases(*safe_keys),
+            _evaluate_aliases(compliant_frame, safe_keys),
             output_names,
         )
 
 
 class DepthTrackingGroupBy(
-    ParseKeysGroupBy[CompliantFrameT, DepthTrackingExprT_contra],
-    Protocol38[CompliantFrameT, DepthTrackingExprT_contra, NativeAggregationT_co],
+    ParseKeysGroupBy[CompliantFrameT_co, DepthTrackingExprT_contra],
+    Protocol[CompliantFrameT_co, DepthTrackingExprT_contra, NativeAggregationT_co],
 ):
     """`CompliantGroupBy` variant, deals with `Eager` and other backends that utilize `CompliantExpr._depth`."""
 
@@ -195,39 +180,9 @@ class DepthTrackingGroupBy(
 
 
 class EagerGroupBy(
-    DepthTrackingGroupBy[CompliantDataFrameT, EagerExprT_contra, NativeAggregationT_co],
-    DataFrameGroupBy[CompliantDataFrameT, EagerExprT_contra],
-    Protocol38[CompliantDataFrameT, EagerExprT_contra, NativeAggregationT_co],
+    DepthTrackingGroupBy[
+        CompliantDataFrameT_co, EagerExprT_contra, NativeAggregationT_co
+    ],
+    DataFrameGroupBy[CompliantDataFrameT_co, EagerExprT_contra],
+    Protocol[CompliantDataFrameT_co, EagerExprT_contra, NativeAggregationT_co],
 ): ...
-
-
-class LazyGroupBy(
-    ParseKeysGroupBy[CompliantLazyFrameT, LazyExprT_contra],
-    CompliantGroupBy[CompliantLazyFrameT, LazyExprT_contra],
-    Protocol38[CompliantLazyFrameT, LazyExprT_contra, NativeExprT_co],
-):
-    _keys: list[str]
-    _output_key_names: list[str]
-
-    def _evaluate_expr(self, expr: LazyExprT_contra, /) -> Iterator[NativeExprT_co]:
-        output_names = expr._evaluate_output_names(self.compliant)
-        aliases = (
-            expr._alias_output_names(output_names)
-            if expr._alias_output_names
-            else output_names
-        )
-        native_exprs = expr(self.compliant)
-        if expr._is_multi_output_unnamed():
-            exclude = {*self._keys, *self._output_key_names}
-            for native_expr, name, alias in zip(native_exprs, output_names, aliases):
-                if name not in exclude:
-                    yield expr._alias_native(native_expr, alias)
-        else:
-            for native_expr, alias in zip(native_exprs, aliases):
-                yield expr._alias_native(native_expr, alias)
-
-    def _evaluate_exprs(
-        self, exprs: Iterable[LazyExprT_contra], /
-    ) -> Iterator[NativeExprT_co]:
-        for expr in exprs:
-            yield from self._evaluate_expr(expr)
